\documentclass{article}

%{{{ packages
\usepackage{verbatim}
%
\makeatletter
\def\itverb{%
\def\verbatim@font{\ttfamily\slshape
                   \let\do\do@noligs
                   \verbatim@nolig@list}%
}%itverb
\makeatother
%
\newenvironment{code}{\small\verbatim}{\endverbatim\normalsize}
\newenvironment{pseudocode}{\small\it\verbatim}{\endverbatim\normalsize}
\newenvironment{session}{\small\itverb\verbatim}{\endverbatim\normalsize}

\usepackage{latexsym,amssymb}
\usepackage[dvips]{graphicx}
%}}}

\author{{\sc Wolfram Kahl}}
\title{ICFP Contest 2001}

%{{{ Settings
\topmargin 0pt
\setlength{\voffset}{-15mm}
\setlength{\hoffset}{-9mm}
\setlength{\oddsidemargin}{0mm}
\setlength{\evensidemargin}{-5.5mm}
\setlength{\textwidth}{172mm}
\setlength{\textheight}{253mm}

\pagestyle{headings}
%%\pagestyle{empty}

\parindent0pt
\parskip4pt
%}}}

\begin{document}
\maketitle

\tableofcontents

%{{{ \datespecial
\edef\dateps{{! userdict begin /bop-hook {gsave 240 30 translate
/Helvetic-Narrow findfont 9 scalefont setfont
0 0 moveto (\number\year-\number\month-\number\day) show
10 0 rmoveto (\number\time) show
30 0 rmoveto
dup 1 add 4 string cvs show
grestore} def end}}
\def\datespecial{\expandafter\special\dateps}
\datespecial
%}}}

\centerline{\texttt{http://cristal.inria.fr/ICFP2001/prog-contest/task.ps}}

\begin{code}
import qualified Char
import System
import Random
import Decks
import Parsing
import STFun
\end{code}

%{{{ \section{Data Structure}
\section{Data Structure}

%{{{ data Tag
\begin{code}
type Colour = Char
colours = "rgbcmykw"

type UL = Int

type Size = Int
sizes :: [Size]
sizes = [0 .. 9]

data Tag = U
         | Col Colour
         | Size Size
         | B
         | EM
         | I
         | PL
         | S
         | TT

tags = [U,B,EM,I,PL,S,TT] ++ map Col colours ++ map Size sizes

asc = filter (`notElem` "<>") [Char.chr 33 .. Char.chr 126]
white = " \n\r\t"
chars = asc ++ white
chars_len = length chars - 1

instance Show Tag where
  showsPrec _ U = ('U' :)
  showsPrec _ (Col c) = (c :)
  showsPrec _ (Size i) = shows i
  showsPrec _ B = ('B' :)
  showsPrec _ EM = ('E' :) . ('M' :)
  showsPrec _ I = ('I' :)
  showsPrec _ PL = ('P' :) . ('L' :)
  showsPrec _ S = ('S' :)
  showsPrec _ TT = ('T' :) . ('T' :)

instance Read Tag where
  readsPrec _ [] = []
  readsPrec _ (c : cs) =
   case c of
    'B' -> [(B,cs)]
    'I' -> [(I,cs)]
    'S' -> [(S,cs)]
    'U' -> [(U,cs)]
    '0' -> [(Size 0,cs)]
    '1' -> [(Size 1,cs)]
    '2' -> [(Size 2,cs)]
    '3' -> [(Size 3,cs)]
    '4' -> [(Size 4,cs)]
    '5' -> [(Size 5,cs)]
    '6' -> [(Size 6,cs)]
    '7' -> [(Size 7,cs)]
    '8' -> [(Size 8,cs)]
    '9' -> [(Size 9,cs)]
    'r' -> [(Col c,cs)]
    'g' -> [(Col c,cs)]
    'b' -> [(Col c,cs)]
    'c' -> [(Col c,cs)]
    'm' -> [(Col c,cs)]
    'y' -> [(Col c,cs)]
    'k' -> [(Col c,cs)]
    'w' -> [(Col c,cs)]
    'E' -> case cs of
            ('M' : cs') -> [(EM,cs')]
            _ -> []
    'P' -> case cs of
            ('L' : cs') -> [(PL,cs')]
            _ -> []
    'T' -> case cs of
            ('T' : cs') -> [(TT,cs')]
            _ -> []
    _   -> []
\end{code}
%}}}

%{{{ data Doc
\begin{code}
tagStart tag = ('<' :) . shows tag . ('>' :)
tagEnd   tag = ('<' :) . ('/' :) . shows tag . ('>' :)

data Doc = Doc (FEDeck String (Tag,Doc))

instance Show Doc where
  showsPrec i (Doc d) = feFold fe ff d
    where fe r1 (tag,d') r2 = r1 . tagStart tag . showsPrec i d' .
                                   tagEnd   tag . r2
          ff s = (s ++)

readsTag :: ReadS (Bool, Tag)
readsTag ('<' : s) =
  case s of
    ('/' : s1) -> readsTag' False s1
    _          -> readsTag' True  s
readsTag _ = []


readsTag' b s = case reads $ dropWhile Char.isSpace s of
  ((tag, s1) : _) -> case dropWhile ('>' /=) s1 of
                       _ : s2 -> [((b,tag),s2)]
                       _ -> []
  _ -> []

readsText s = [span ('<' /=) s]

readsStartTag :: ReadS Tag
readsStartTag = readsTag <?>
  (\ (b,tag) -> if b then Just tag else Nothing)

readsEndTag :: Tag -> ReadS ()
readsEndTag tag = readsTag <?>
  (\ (b,tag) -> if b then Nothing else Just ())

readsTagDoc :: ReadS (Tag,Doc)
readsTagDoc = readsStartTag `pThen` \ tag ->
              (\ d -> (tag,d)) <$>
              (readsDoc `pThen` \ doc ->
               readsEndTag tag `pThen` \ _ -> pSucceed doc)

readsDoc :: ReadS Doc
readsDoc = readsText `pThen` \ f ->
           Doc <$>
           ((readsTagDoc `pThen` \ p ->
             readsDoc `pThen` \ (Doc d) ->
             pSucceed (Decks.cons f p d)
            )
            <||> pSucceed (singletonF f))

instance Read Doc where
  readsPrec _ = readsDoc
\end{code}
%}}}

%{{{ generateRandomDoc
\begin{code}
swapPair (x,y) = (y,x)

randomRSTF :: (Random a, RandomGen g) => (a, a) -> STFun g a
randomRSTF bds = STFun (swapPair . randomR bds)

rChar :: RandomGen g => STFun g Char
rChar = do i <- randomRSTF (0, chars_len)
           return (chars !! i)

rLen :: RandomGen g => Int -> STFun g Int
rLen i = randomRSTF (0,i)

rText :: RandomGen g => STFun g String
rText = do len <- rLen 4
           mapM (const rChar) [1 .. len]

rTag :: RandomGen g => STFun g Tag
rTag = do i <- rLen 24
          return (tags !! i)

rDeck w n = do t <- rText
               if n < 1 || w < 1
                then return (singletonF t)
                else do d <- rDoc (n - 1)
                        tag <- rTag
                        r <- rDeck (w - 1) n
                        return (Decks.cons t (tag,d) r)

rDoc n = do depth <- rLen n
            width <- rLen 10
            d <- rDeck width n
            return $ Doc d

generateRandomDoc =
 do args <- getArgs
    let n = case args of
              [] -> 6
              (a : _) -> case reads a of
                 ((i,_):_) -> i :: Int
                 _ -> 6
    rand <- newStdGen
    print $ runSTFun (rDoc n) rand
\end{code}
%}}}

%}}}

%{{{ \section{Optimize}
\section{Optimize}

%{{{ data Context
\begin{code}
data Context = Context
  {c_b    :: Bool
  ,c_em   :: Bool
  ,c_i    :: Bool
  ,c_s    :: Bool
  ,c_tt   :: Bool
  ,c_u    :: Int
  ,c_size :: Maybe Int
  ,c_col  :: Maybe Char
  } deriving Eq

c_init = Context
  {c_b = False
  ,c_em = False
  ,c_i = False
  ,c_s = False
  ,c_tt = False
  ,c_u = 0
  ,c_size = Nothing
  ,c_col = Nothing
  }
\end{code}
%}}}

\verb|c_tag tag c| is the context inside a \verb|tag|-element
if \verb|c| is the context outside.

%{{{ c_tag
\begin{code}
c_tag :: Tag -> Context -> Context
c_tag B  c = c {c_b = True} 
c_tag EM c = if c_s c then c
             else c {c_em = not $ c_em c} 
c_tag I  c = c {c_i = True}
c_tag PL c = c {c_s  = False
               ,c_em = False
               ,c_i  = False
               ,c_b  = False
               ,c_tt = False
               ,c_u  = 0
               }
c_tag S  c = c {c_s = True, c_em = False}
c_tag TT c = c {c_tt = True}
c_tag U  c = let u = c_u c
             in if u >= 3 then c else c {c_u = u + 1}
c_tag (Size i) c = c {c_size = Just i}
c_tag (Col co) c = c {c_col  = Just co}
\end{code}
%}}}

\verb|c_implies c tag| holds iff \verb|c_tag tag c == c|.

\begin{code}
c_implies c tag = c_tag tag c == c
\end{code}

\begin{code}
isWSpace c = c `elem` white

wCompress1 [] = []
wCompress1 (c:cs) | isWSpace c  =  c : wCompress1 (dropWhile isWSpace cs)
                  | otherwise   =  c : wCompress1 cs
\end{code}

\begin{code}
opt1doc c (Doc d) = Doc $ opt1deck c d
opt1deck c = feFold fe ff
 where
  ff = singletonF . (if c_tt c then id else wCompress1)
  fe d1 (tag, doc@(Doc d)) d2 =
       if c_implies c tag
       then d1 `cat` (opt1deck c d `cat` d2)
       else ecat d1 (tag, opt1doc (c_tag tag c) doc) d2
\end{code}
%}}}

\begin{code}
mirror s = case readsDoc s of
            ((d,_):_) -> show (opt1doc c_init d)
            _ -> []

-- main = generateRandomDoc
main = interact mirror
\end{code}

\section{Auxiliary Material}

\input{Parsing.lhs}

\end{document}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
